<?php

declare(strict_types=1);

namespace mirzaev\calculator\models;

use mirzaev\calculator\models\settings_model as settings;
use mirzaev\calculator\models\metals_model as metals;

use exception;

/**
 * Модель калькуляторов
 *
 * @package mirzaev\calculator\models
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
final class calculators_model extends core
{

    /**
     * Расчёт стоимости переработки за 1 тонну
     *
     * @param string $complexity Сложность
     * @param float $lenght Толщина
     * @param array &$errors Журнал ошибок
     *
     * @return float Стоимость переработки за 1 тонну (руб)
     */
    public static function reprocessing(string $complexity, float $lenght, array &$errors = []): ?float
    {
        try {
            return (float) match (true) {
                $lenght > 0 && $lenght < 4 => settings::read('reprocessing_' . $complexity . '_1_3', $errors),
                $lenght > 3 && $lenght < 7 => settings::read('reprocessing_' . $complexity . '_4_6', $errors),
                $lenght > 6 && $lenght < 11 => settings::read('reprocessing_' . $complexity . '_7_10', $errors),
                default => settings::read('reprocessing_' . $complexity . '_10', $errors),
            }
                ?? 0.0;
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[] = $e->getMessage() . ':' . $e->getLine();
        }

        return 0.0;
    }

    /**
     * Расчёт коэффициента сложности детали
     *
     * Влияет на скорость реза
     *
     * @param string $complexity Сложность детали (easy, medium, hard)
     * @param int|null $area Площадь детали
     * @param float|null $lenght Толщина детали
     * @param array &$errors Журнал ошибок
     *
     * @return float Коэффициент
     *
     * @todo
     * 1. Коэффициент исходя из типа газа (баллоны)
     */
    public static function coefficient(string $complexity, ?int $area = null, ?float $lenght = null, array &$errors = []): float|false
    {
        try {
            // Коэффициент полученный исходя из сложности детали
            $coefficient = settings::read("coefficient_complexity_$complexity", $errors) ?? throw new exception("Не найдено: coefficient_complexity_$complexity");

            if (
                isset($area)
                && ($area <= (settings::read('coefficient_area_less', $errors) ?? throw new exception("Не найдено: coefficient_area_less"))
                    || $area >= (settings::read('coefficient_area_more', $errors) ?? throw new exception("Не найдено: coefficient_area_more")))
            ) {
                // Площадь детали не более и не менее заданных в базе данных размеров

                // Прибавление коэффициента исходя из площади детали
                $coefficient += settings::read('coefficient_area_degree', $errors) ?? throw new exception("Не найдено: coefficient_area_degree");
            }

            if (
                isset($lenght)
                && ($lenght <= (settings::read('coefficient_lenght_less', $errors) ?? throw new exception("Не найдено: coefficient_lenght_less"))
                    || $lenght >= (settings::read('coefficient_lenght_more', $errors) ?? throw new exception("Не найдено: coefficient_lenght_more")))
            ) {
                // Толщина детали не более и не менее заданных в базе данных размеров

                // Прибавление коэффициента исходя из толщины детали
                // $coefficient += settings::read('coefficient_lenght_degree', $errors) ?? throw new exception("Не найдено: coefficient_lenght_degree");
                // $coefficient -= settings::read('coefficient_lenght_degree', $errors) ?? throw new exception("Не найдено: coefficient_lenght_degree");
            }

            return (float) $coefficient;
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[] = [
                'text' => $e->getMessage(),
                'file' => $e->getFile(),
                'line' => $e->getLine(),
                'stack' => $e->getTrace()
            ];
        }

        return false;
    }

    /**
     * Расчёт стоимости работы сотрудника
     *
     * @param string $position Должность (manager, engineer, operator...)
     * @param string $complexity Сложность детали (easy, medium, hard)
     * @param int $company Для юридического лица? Иначе как для физического
     * @param array &$errors Журнал ошибок
     *
     * @return int Стоимость (руб)
     */
    public static function employee(string $position, string $complexity, bool $company = false, array &$errors = []): array|bool
    {
        try {
            // Инициализация категории работы для поиска данных
            $type = $company ? 'entity' : 'individual';

            // Инициализация всех времён работы (ч)
            $min = settings::read($position . '_' . $type . '_min', $errors) ?? throw new exception('Не найдено: ' . $position . '_' . $type . '_min');
            $max = settings::read($position . '_' . $type . '_max', $errors) ?? throw new exception('Не найдено: ' . $position . '_' . $type . '_max');
            $average = ($min + $max) / 2;

            // Инициализация времени работы
            $time = (float) match ($complexity) {
                'easy' => $min,
                'medium' => $average,
                'hard' => $max,
                default => throw new exception('Неизвестный тип сложности детали')
            };

            // Инициализация цены за час
            $hour = settings::read($position . '_' . $type . '_hour', $errors) ?? throw new exception('Не найдено: ' . $position . '_' . $type . '_hour');

            return [
                'time' => (float) $time,
                'hour' => (float) $hour,
                'cost' => $time * $hour
            ];
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[] = [
                'text' => $e->getMessage(),
                'file' => $e->getFile(),
                'line' => $e->getLine(),
                'stack' => $e->getTrace()
            ];
        }

        return false;
    }

    /**
     * Расчёт калькулятора лазерной резки
     *
     * @param bool|int|string|null $company Юридическое лицо? Или физическое лицо?
     * @param string|null $complexity Сложность ('easy', 'medium', 'hard')
     * @param int|string|null $width Высота (мм)
     * @param int|string|null $height Ширина (мм)
     * @param int|string|null $length Длина (мм)
     * @param int|string|null $amount Количество
     * @param string|null $type Тип металла
     * @param string|null $mark Марка металла
     * @param bool|null $out Наш металл?
     * @param int|string|null $holes Количество отверстий
     * @param int|string|null $diameter Диаметр отверстий (мм)
     * @param string|null $discount Скидка менеджера
     * @param array &$errors Журнал ошибок
     *
     * @return array|bool Аккаунт, если удалось аутентифицироваться
     *
     * @todo
     * 18. 1200 баллон кислород расход 1 баллон на 3 листа 4ки
     * 22. Выводится результат даже если не аутентифицирован (проблема с куки)
     * 29. Амортизация станка
     * 30. Аренда помещения
     * 31. Расходники
     * 37. Проверка на то, что загружен минимум 1 калькулятор
     */
    public static function laser(
        bool|int|string|null $company = null,
        ?string $complexity = null,
        int|string|null $width = null,
        int|string|null $height = null,
        float|string|null $length = null,
        float|string|null $amount = null,
        ?string $type = null,
        ?string $mark = null,
        ?bool $our = null,
        int|string|null $holes = null,
        float|string|null $diameter = null,
        float|string|null $discount = null,
        array &$errors = []
    ): array {
        // Инициализация журнала ошибок
        $errors['laser'] ?? $errors['laser'] = [];
        $i = count($errors['laser']);
        $errors['laser'][$i] = [];
        $errors = $errors['laser'][$i];

        // Инициализация переменных для буфера вывода
        $machines = $managers = $engineers = $operators = $handymans = $other = [];

        try {
            // Инициализация значений по умолчанию
            $company = $company ?? throw new exception('Не передан параметр company');
            $complexity = $complexity ?? throw new exception('Не передан параметр complexity');
            $width = (int) $width ?? throw new exception('Не передан параметр width');
            $height = (int) $height ?? throw new exception('Не передан параметр height');
            $length = (float) $length ?? throw new exception('Не передан параметр length');
            $amount = (float) $amount ?? throw new exception('Не передан параметр amount');
            $type = $type ?? throw new exception('Не передан параметр type');
            $mark = $mark ?? throw new exception('Не передан параметр mark');
            $our = (bool) $our ?? true;
            $holes = (int) $holes ?? throw new exception('Не передан параметр holes');
            $diameter = (float) $diameter ?? throw new exception('Не передан параметр diameter');
            $discount = (float) $discount ?? throw new exception('Не передан параметр discount');

            if ($width <= 0 || $height <= 0 || $length <= 0) {
                // Неподходящие для выполнения значения

                throw new exception('Передан нулевой размер одной из сторон заготовки');
            }

            // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            // $gas = 'oxygen';
            $gas = 'air';

            // Инициализация станка для буфера вывода
            $machine = [
                'amount' => $amount
            ];

            // Инициализация буфера остальных вычислений для буфера вывода
            $other = [
                'additive' => (float) settings::read('additive', $errors) ?? throw new exception('Не найдено: additive'),
                'discount' => $discount
            ];

            // Инициализация прибавок к наценке
            $increase_300_3000 = (float) settings::read('additive_increase_300_3000', $errors) ?? throw new exception('Не найдено: additive_increase_300_3000');

            if ($amount >= 300) {
                // Количество заказанных деталей равно или более чем 300 шт.

                for ($i = 0; $i <= $amount; ++$i) {
                    // Перебор по количеству изготавливаемых деталей

                    // Количество заказанных деталей равно или более чем 300 шт. и равно или менее чем 3000 шт.
                    if ($amount <= 3000) $other['additive'] += $increase_300_3000;
                }
            }

            $other['additive'] = 1.07;

            // Площадь
            $area = $width * $height;

            // Объём
            $volume  = $area * $length;

            // Стоимость киловатта электроэнергии
            $electricity = settings::read('electricity', $errors) ?? throw new exception('Не найдено: electricity');

            // Потребляемая электроэнергия станком (квт/ч)
            $power = settings::read('laser_power', $errors) ?? throw new exception('Не найдено: laser_power');

            // Чтение данных листа металла (почему-то не читает если передать $length = 0.8)
            $list = metals::read($type, $mark, $length, $errors);

            // Проверка
            if (empty($list['ton'])) throw new exception('Не удалось найти цену за тонну');
            if (empty($list['width'])) throw new exception('Не удалось найти ширину листа');
            if (empty($list['height'])) throw new exception('Не удалось найти высоту листа');
            if ((int) $list['width'] < $width) throw new exception('Ширина детали крупнее чем у исходного листа');
            if ((int) $list['height'] < $height) throw new exception('Высота детали крупнее чем у исходного листа');

            // Объём листа металла
            $list['volume'] = $list['width'] * $list['height'] * $length;

            // Инициализация характеристик металла
            $metal = [
                'cut' => metals::cut($type, $gas, $length, $errors),
                'weight' => metals::kg($type, $errors) * $length,
                'cost' => $list['ton'] / 1000 // Цена за килограмм
            ];

            if ($metal['cut'] <= 0 || $metal['weight'] <= 0 || $metal['cost'] <= 0) {
                // Неподходящие для выполнения значения

                throw new exception('Ошибка при вычислении характеристик металла');
            }

            if ($holes == 0 ?? $diameter == 0) {
                // Не переданы данные об отверстиях

                // Длина реза (мм)
                $cutting = ($width * 2 + $height * 2) * self::coefficient($complexity, $area, $length, $errors);
            } else {
                // Переданы данные об отверстиях

                // Длина реза (мм)
                $cutting = (3.1416 * $diameter * $holes + $width * 2 + $height * 2) * self::coefficient($complexity, $area, $length, $errors);
            }

            $metal['cut'] = 65;

            // Время реза одной детали (c)
            $time = $cutting / $metal['cut'];

            // Стоимость реза в час (руб)
            $hour = $electricity * $power;

            // Стоимость электроэнергии за работу станка
            $machine['electricity'] = $time * $amount / 60 / 60 * $hour;

            // Вес металла (кг)
            $weight = $width * $height * $amount / 1000000 * $metal['weight'];

            // Стоимость переработки (руб)
            $other['reprocessing'] = self::reprocessing($complexity, $length, $errors) / 1000 * $weight;

            // (Наш металл) Стоимость металла (руб)
            if ($our) $machine['metal'] = $weight * $metal['cost'];

            // Инициализация времени загрузки металла на станок
            $load = (float) settings::read('handyman_metal_load_time') ?? throw new exception('Не найдено: handyman_metal_load_time');

            // Инициализация стоимости часа
            $hour = (float) settings::read('handyman_hour') ?? throw new exception('Не найдено: handyman_hour');

            // Инициализация количества разнорабочих
            $workers = settings::read('laser_handymans_amount') ?? throw new exception('Не найдено: laser_handymans_amount');

            // Вычисление количества листов
            for ($lists = 0; ++$lists * $list['volume'] < $volume * $amount;);

            // Вычисление времени работы разнорабочих (ч)
            $work = $lists * $load / 60;

            // Инициализация итератора
            $i = 0;

            while (++$i <= $workers) {
                // Перебор разнорабочих

                // Инициализация разнорабочего
                $handymans[$i] = [
                    'time' => $work, // Время работы (ч)
                    'hour' => $hour, // Стоимость работы в час (руб)
                    'cost' => $work * $hour // Стоимость работы
                ];
            }

            if ($company) {
                // Юридическое лицо

                // Инициализация менеджеров
                $managers = [static::employee('manager', $complexity, true, errors: $errors)];

                // Инициализация инженеров
                $engineers = [static::employee('engineer', $complexity, true, errors: $errors)];

                // Инициализация стоимости работы оператора
                $operator = static::employee('operator', $complexity, true, errors: $errors);

                // Инициализация стоимости 1 часа работы оператора
                $hour = settings::read('operator_entity_hour') ?? throw new exception('Не найдено: operator_entity_hour');

                // Инициализация операторов
                $operators = [
                    [
                        'time' => [
                            'design' => $operator['time'],
                            'machine' => ($time * $amount / 60 / 60) + $work
                        ],
                        'hour' => $operator['hour']
                    ]
                ];
            } else {
                // Физическое лицо

                // Инициализация менеджеров
                $managers = [static::employee('manager', $complexity, errors: $errors)];

                // Инициализация инженеров
                $engineers = [static::employee('engineer', $complexity, errors: $errors)];

                // Инициализация данных оператора
                $operator = static::employee('operator', $complexity, errors: $errors);

                // Инициализация стоимости 1 часа работы оператора
                $hour = settings::read('operator_individual_hour') ?? throw new exception('Не найдено: operator_individual_hour');

                // Инициализация операторов
                $operators = [
                    [
                        'time' => [
                            'design' => $operator['time'],
                            'machine' => ($time * $amount / 60 / 60) + $work
                        ],
                        'hour' => $operator['hour']
                    ]
                ];
            }

            // Инициализация станков
            $machines = [
                $machine
            ];
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[] = [
                'text' => $e->getMessage(),
                'file' => $e->getFile(),
                'line' => $e->getLine(),
                'stack' => $e->getTrace()
            ];
        }

        return [$machines, $managers, $engineers, $operators, $handymans, $other];
    }
}
