<?php

declare(strict_types=1);

namespace mirzaev\calculator\models;

use pdo;
use exception;

/**
 * Модель настроек
 *
 * @package mirzaev\calculator\models
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
final class settings_model extends core
{
    /**
     * Записать настройку
     *
     * @param string $name Название параметра
     * @param int|float|string|bool $value Значение параметра
     * @param array &$errors Журнал ошибок
     *
     * @return bool Статус записи
     */
    public static function write(string $name, int|float|string|bool $value, array &$errors = []): bool
    {
        try {

            // Инициализация запроса
            $request = static::$db->prepare("SELECT MAX(`id`) FROM `settings`");

            // Отправка запроса
            $request->execute();

            // Получение ответа
            $id = (int) $request->fetchColumn();

            // Инициализация запроса
            $request = static::$db->prepare("UPDATE `settings` SET `$name` = :value WHERE `id` = $id");

            // Параметры запроса
            $params = [
                ":value" => $value
            ];

            // Отправка запроса
            $request->execute($params);

            // Получение ответа
            return $request->fetch(pdo::FETCH_ASSOC) !== false;
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[]= [
                'text' => $e->getMessage(),
                'file' => $e->getFile(),
                'line' => $e->getLine(),
                'stack' => $e->getTrace()
            ];
        }

        return false;
    }

    /**
     * Прочитать настройку
     *
     * @param string $name Название параметра
     * @param array &$errors Журнал ошибок
     *
     * @return int|float|string|array|null Значение параметра
     */
    public static function read(string $name = '*', array &$errors = []): int|float|string|array|null
    {
        try {
            // Инициализация запроса
            $request = static::$db->prepare("SELECT $name FROM `settings` WHERE `id` = (SELECT MAX(`id`) FROM `settings`) ORDER BY `id` DESC LIMIT 1");

            // Отправка запроса
            $request->execute();

            // Генерация ответа
            $response = $name === '*' ? $request->fetch(pdo::FETCH_ASSOC) : $request->fetchColumn();

            return $response === false ? null : $response;
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[] = [
                'text' => $e->getMessage(),
                'file' => $e->getFile(),
                'line' => $e->getLine(),
                'stack' => $e->getTrace()
            ];
        }

        return null;
    }

    /**
     * Прочитать названия настроек
     *
     * @param int $limit Максимальное ограничение
     * @param array &$errors Журнал ошибок
     *
     * @return array|null Значение параметра
     */
    public static function names(int $limit = 5000, array &$errors = []): ?array
    {
        try {
            // Инициализация запроса
            $request = static::$db->prepare("SELECT TABLE_NAME AS 'settings' FROM INFORMATION_SCHEMA.COLUMNS WHERE `table_name` = 'settings'");

            // Отправка запроса
            $request->execute();

            // Генерация ответа
            $response = $request->fetchColumn();

            return $response === false ? null : (array) $response;
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[]= [
                'text' => $e->getMessage(),
                'file' => $e->getFile(),
                'line' => $e->getLine(),
                'stack' => $e->getTrace()
            ];
        }

        return null;
    }

    /**
     * Фильтровать настройки
     *
     * Удаляет из массива ненужные данные (id, create...)
     *
     * @param array $settings Настройки
     * @param array &$errors Журнал ошибок
     *
     * @return array Настройки
     */
    public static function filter(array $settings, array &$errors = []): array
    {
        try {
            // Инициализация буфера вывода
            $buffer = [];

            foreach ($settings as $setting => $value) {
                // Перебор настроек

                if ($setting === 'id') continue;
                if ($setting === 'create') continue;

                $buffer[$setting] = $value;
            }

            return $buffer ?? $settings;
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[]= [
                'text' => $e->getMessage(),
                'file' => $e->getFile(),
                'line' => $e->getLine(),
                'stack' => $e->getTrace()
            ];
        }

        return $settings;
    }

    /**
     * Перевести настройки на русский язык
     *
     * @param array $settings Настройки
     * @param array &$errors Журнал ошибок
     *
     * @return array Настройки
     */
    public static function translate(array $settings, array &$errors = []): array
    {
        try {
            // Инициализация буфера вывода
            $buffer = [];

            foreach ($settings as $setting => $value) {
                // Перебор настроек

                $buffer[] = [
                    'russian' => match ($setting) {
                        'default_buyer' => 'Покупатель по умолч.',
                        'default_complexity' => 'Сложность по умолч.',
                        'default_width' => 'Ширина по умолч. (мм)',
                        'default_height' => 'Высота по умолч. (мм)',
                        'default_length' => 'Толщина по умолч. (мм)',
                        'default_amount' => 'Количество  по умолч. (шт)',
                        'default_type' => 'Тип металла по умолч.',
                        'default_holes' => 'Отверстия по умол. (шт)',
                        'default_diameter' => 'Диаметр отверстий по умол. (мм)',
                        'electricity' => 'Стоимость 1квт',
                        'laser_power' => 'Потребление лазер. станка (квт)',
                        'manager_individual_min' => 'Менеджер для физ. лица мин. (ч)',
                        'manager_individual_max' => 'Менеджер для физ. лица макс. (ч)',
                        'manager_entity_min' => 'Менеджер для юр. лица мин. (ч)',
                        'manager_entity_max' => 'Менеджер для юр. лица макс. (ч)',
                        'manager_entity_hour' => 'Менеджер для юр. лица (руб/час)',
                        'manager_individual_hour' => 'Менеджер для физ. лица (руб/час)',
                        'engineer_entity_min' => 'Инженер для юр. лица мин. (ч)',
                        'engineer_entity_max' => 'Инженер для юр. лица макс. (ч)',
                        'engineer_entity_hour' => 'Инженер для юр. лица (руб/час)',
                        'engineer_individual_min' => 'Инженер для физ. лица мин. (ч)',
                        'engineer_individual_max' => 'Инженер для физ. лица макс. (ч)',
                        'engineer_individual_hour' => 'Инженер для физ. лица (руб/час)',
                        'operator_entity_min' => 'Оператор для юр. лица мин. (ч)',
                        'operator_entity_max' => 'Оператор для юр. лица макс. (ч)',
                        'operator_individual_min' => 'Оператор для физ. лица мин. (ч)',
                        'operator_individual_max' => 'Оператор для физ. лица макс. (ч)',
                        'operator_entity_hour' => 'Оператор для юр. лица (руб/час)',
                        'operator_individual_hour' => 'Оператор для физ. лица (руб/час)',
                        'coefficient_area_less' => 'Штраф если пл. меньше (мм)',
                        'coefficient_area_more' => 'Штраф если пл. больше (мм)',
                        'coefficient_area_degree' => 'Коэфф. для штрафа по площади',
                        'coefficient_complexity_easy' => 'Коэфф. низкой слож.',
                        'coefficient_complexity_medium' => 'Коэфф. средней слож.',
                        'coefficient_complexity_hard' => 'Коэфф. высокой слож.',
                        'list_width' => 'Ширина листа (мм)',
                        'list_height' => 'Высота листа (мм)',
                        'metal_cost_easy_1_3' => 'Тонна 1-3мм низкая слож. (руб)',
                        'metal_cost_medium_1_3' => 'Тонна 1-3мм средняя слож. (руб)',
                        'metal_cost_hard_1_3' => 'Тонна 1-3мм высокая слож. (руб)',
                        'metal_cost_easy_4_6' => 'Тонна 4-6мм низкая слож. (руб)',
                        'metal_cost_medium_4_6' => 'Тонна 4-6мм средняя слож. (руб)',
                        'metal_cost_hard_4_6' => 'Тонна 4-6мм высокая слож. (руб)',
                        'metal_cost_easy_7_10' => 'Тонна 7-10мм низкая слож. (руб)',
                        'metal_cost_medium_7_10' => 'Тонна 7-10мм средняя слож. (руб)',
                        'metal_cost_hard_7_10' => 'Тонна 7-10мм высокая слож. (руб)',
                        'metal_cost_easy_10' => 'Тонна >10мм низкая слож. (руб)',
                        'metal_cost_medium_10' => 'Тонна >10мм средняя слож. (руб)',
                        'metal_cost_hard_10' => 'Тонна >10мм высокая слож. (руб)',
                        'additive' => 'Коэфф. наценки',
                        'handyman_hour' => 'Разнорабочий (руб/ч)',
                        'handyman_metal_load_time' => 'Время установки металла (мин)',
                        'laser_handymans_amount' => 'Разнорабочие для лазерного станка',
                        default => $setting
                    },
                    'english' => $setting,
                    'value' => $value
                ];
            }

            return $buffer ?? $settings;
        } catch (exception $e) {
            // Запись в журнал ошибок
            $errors[]= [
                'text' => $e->getMessage(),
                'file' => $e->getFile(),
                'line' => $e->getLine(),
                'stack' => $e->getTrace()
            ];
        }

        return $settings;
    }
}
